from django.test import TestCase
from aws_message.sqs import SQSQueue


SETTINGS = {
    'TOPIC_ARN': 'arn:aws:sqs:mock:123456789012:ww-wwww-1',
    'KEY_ID': 'XXXXXXXXXXXXXXXX',
    'KEY': 'YYYYYYYYYYYYYYYYYYYYYYYY',
    'VISIBILITY_TIMEOUT': 10,
    'MESSAGE_GATHER_SIZE': 10,
    'VALIDATE_SNS_SIGNATURE': True,
    'PAYLOAD_SETTINGS': {}
}


class TestSQSQueue(TestCase):

    def test_get_queue(self):
        pass



from django.test import TestCase, override_settings
from django.conf import settings
from aws_message.message import Message
from aws_message.crypto import CryptoException

TEST_MESSAGES = [
    {'abc': 'def'},
    'abcdef 123456789',
    '[a, b, c]',
    'TopicArn: nope',
    ['a', 'b', 'c'],
]

TEST_MSG_SNS = {
    'SignatureVersion': '2',
    'Timestamp': '2018-08-22T04:00:56.843Z',
    'Signature': '...',
    'SigningCertURL': '',
    'MessageId': '11111111-0000-1111-2222-555555555555',
    'Message': (
        '{"EventID":"00000000-1111-2222-3333-444444444444","Href":'
        '"...","EventDate":"2018-08-21T21:00:56.832068-07:00","Previous":'
        '{"CurrentEnrollment":19,"Status":"open"},"Current":'
        '{"CurrentEnrollment":20,"Status":"closed"}}'),
    'UnsubscribeURL': '',
    'Type': 'Notification',
    'TopicArn': 'arn:aws:sns:us-east-1:111111111111:test-topic',
    'Subject': 'UW Event',
}

TEST_MSG_SNS_B64 = {
    'SignatureVersion': '1',
    'Timestamp': '2018-08-22T04:00:56.843Z',
    'Signature': '...',
    'SigningCertURL': '',
    'MessageId': '11111111-0000-1111-2222-555555555555',
    'Message': (
        'eyJFdmVudElEIjoiMDAwMDAwMDAtMTExMS0yMjIyLTMzMzMtNDQ0NDQ0NDQ0NDQ0'
        'IiwiSHJlZiI6Ii4uLiIsIkV2ZW50RGF0ZSI6IjIwMTgtMDgtMjFUMjE6MDA6NTYu'
        'ODMyMDY4LTA3OjAwIiwiUHJldmlvdXMiOnsiQ3VycmVudEVucm9sbG1lbnQiOjE5'
        'LCJTdGF0dXMiOiJvcGVuIn0sIkN1cnJlbnQiOnsiQ3VycmVudEVucm9sbG1lbnQi'
        'OjIwLCJTdGF0dXMiOiJjbG9zZWQifX0='),
    'UnsubscribeURL': '',
    'Type': 'Notification',
    'TopicArn': 'arn:aws:sns:us-east-1:111111111111:test-topic',
    'Subject': 'UW Event',
}


class TestMessageExtract(TestCase):
    @override_settings(AWS_SQS={'TEST': {}})
    def test_extract_inner_message_str(self):
        for msg in TEST_MESSAGES:
            message = Message(msg, settings.AWS_SQS['TEST'])
            body = message.extract()
            self.assertEquals(body, msg)

    @override_settings(AWS_SQS={'TEST': {}})
    def test_extract_inner_message_sns_base64(self):
        message = Message(TEST_MSG_SNS_B64, settings.AWS_SQS['TEST'])
        body = message.extract()

        self.assertEquals(
            body["EventID"], '00000000-1111-2222-3333-444444444444')
        self.assertEquals(body["Href"], '...')
        self.assertEquals(
            body["EventDate"], '2018-08-21T21:00:56.832068-07:00')

    @override_settings(AWS_SQS={'TEST': {}})
    def test_extract_inner_message_sns_json(self):
        message = Message(TEST_MSG_SNS, settings.AWS_SQS['TEST'])
        body = message.extract()

        self.assertEquals(
            body["EventID"], '00000000-1111-2222-3333-444444444444')
        self.assertEquals(body["Href"], '...')
        self.assertEquals(
            body["EventDate"], '2018-08-21T21:00:56.832068-07:00')


class TestMessageValidate(TestCase):
    @override_settings(AWS_SQS={'TEST': {}})
    def test_validate(self):
        for msg in TEST_MESSAGES:
            message = Message(msg, settings.AWS_SQS['TEST'])
            self.assertEquals(message.validate(), True)

    @override_settings(AWS_SQS={'TEST': {
            'VALIDATE_SNS_SIGNATURE': False,
            'TOPIC_ARN': 'arn:aws:sns:us-east-1:111111111111:test-topic'}},)
    def test_validate_topic_arn(self):
        message = Message(TEST_MSG_SNS, settings.AWS_SQS['TEST'])
        self.assertEquals(message.validate(), True)

    @override_settings(AWS_SQS={'TEST': {
            'VALIDATE_SNS_SIGNATURE': False,
            'TOPIC_ARN': 'arn:aws:sns:us-east-1:111111111111:wrong-topic'}},)
    def test_validate_topic_arn_diff(self):
        message = Message(TEST_MSG_SNS, settings.AWS_SQS['TEST'])
        self.assertEquals(message.validate(), False)

    @override_settings(AWS_SQS={'TEST': {'VALIDATE_SNS_SIGNATURE': True}},
                       AWS_CA_BUNDLE='ca_certs.txt')
    def test_validate_message_invalid_signature(self):
        message = Message(TEST_MSG_SNS, settings.AWS_SQS['TEST'])
        with self.assertRaises(CryptoException) as cm:
            message.validate()
        self.assertEquals(
            'Unknown SNS Signature Version: 2', str(cm.exception))

    @override_settings(AWS_SQS={'TEST': {'VALIDATE_SNS_SIGNATURE': True}},
                       AWS_CA_BUNDLE='ca_certs.txt')
    def test_validate_message_signature(self):
        message = Message(TEST_MSG_SNS_B64, settings.AWS_SQS['TEST'])
        with self.assertRaises(CryptoException) as cm:
            message.validate()
        self.assertEquals(
            'Cannot get certificate None: No host specified.',
            str(cm.exception))



import logging
from django.test import TestCase, override_settings
from aws_message.processor import MessageBodyProcessor, ProcessorException

logger = logging.getLogger(__name__)


class MockProcessor(MessageBodyProcessor):
    def __init__(self, is_encrypted=False):
        super(MockProcessor, self).__init__(
            logger, queue_settings_name='TEST', is_encrypted=is_encrypted)

        self.event_date = None
        self.href = None
        self.current = None

    def process_message_body(self, json_data):
        self.event_date = json_data['EventDate']
        self.href = json_data['Href']
        self.current = json_data['Current']


class MockInvalidProcessor(MockProcessor):
    def validate_message_body(self, message):
        return False


class UnimplementedProcessor(MessageBodyProcessor):
    pass


M1 = {
    "EventID": "00000000-1111-2222-3333-444444444444",
    "Href": "...",
    "EventDate": "2018-08-12T16:39:08.2704415-07:00",
    "Previous": {
        "CurrentEnrollment": 137,
        "CurrentRegistrationPeriod": "2",
        "Status": "open"},
    "Current": {
        "CurrentEnrollment": 138,
        "CurrentRegistrationPeriod": "2",
        "Status": "open"}
}


class TestUnimplementedProcessor(TestCase):
    @override_settings(AWS_SQS={'TEST': {}})
    def test_process(self):
        self.assertRaises(
            TypeError, UnimplementedProcessor, logger,
            queue_settings_name='TEST')


class TestMockProcessor(TestCase):
    @override_settings(AWS_SQS={'TEST': {}})
    def test_process(self):
        processor = MockProcessor()
        valid = processor.process(M1)
        self.assertEqual(
            processor.event_date, "2018-08-12T16:39:08.2704415-07:00")
        self.assertEqual(
            processor.href, "...")
        self.assertIsNotNone(processor.current)

    @override_settings(AWS_SQS={'TEST': {}})
    def test_process_with_validation(self):
        processor = MockInvalidProcessor(is_encrypted=True)
        valid = processor.process(M1)
        self.assertIsNone(processor.event_date)
        self.assertIsNone(processor.href)
        self.assertIsNone(processor.current)

    @override_settings(AWS_SQS={'TEST': {}})
    def test_process_encrypted(self):
        processor = MockProcessor(is_encrypted=True)
        processor.process(M1)
        self.assertEqual(
            processor.event_date, "2018-08-12T16:39:08.2704415-07:00")
        self.assertEqual(
            processor.href, "...")
        self.assertIsNotNone(processor.current)

    @override_settings(AWS_SQS={'TEST': {'VALIDATE_MSG_SIGNATURE': True}})
    def test_process_with_signature_validation(self):
        processor = MockProcessor()
        processor.process(M1)
        self.assertEqual(
            processor.event_date, "2018-08-12T16:39:08.2704415-07:00")
        self.assertEqual(
            processor.href, "...")
        self.assertIsNotNone(processor.current)



from aws_message.crypto import Signature, CryptoException
from base64 import b64decode
import json
import re

re_base64 = re.compile(
    r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|'
    r'[A-Za-z0-9+/]{3}=)?$')


class Message(object):
    def __init__(self, message, settings):
        self._message = message
        self._settings = settings
        self._is_sns = (isinstance(message, dict) and 'TopicArn' in message)

    def validate(self):
        valid = True
        if self._is_sns:
            # SNS message validation
            valid = self._validate_topic_arn()

            if valid and self._settings.get('VALIDATE_SNS_SIGNATURE', True):
                self._validate_sns_signature()

        return valid

    def extract(self):
        message = self._message
        if self._is_sns:
            message = message.get('Message')

        if not isinstance(message, str):
            return message

        if re_base64.match(message):
            message = b64decode(message)

        try:
            return json.loads(message)
        except json.decoder.JSONDecodeError:
            return message

    def _validate_topic_arn(self):
        if (self._settings.get('TOPIC_ARN') and
                self._message['TopicArn'] != self._settings['TOPIC_ARN']):
            return False
        return True

    def _validate_sns_signature(self):
        if self._message['SignatureVersion'] != '1':
            raise CryptoException('Unknown SNS Signature Version: {}'.format(
                self._message['SignatureVersion']))

        sig_conf = {
            'cert': {
                'type': 'url',
                'reference': self._message['SigningCertURL']
            }
        }

        Signature(sig_conf).validate(self._sign_text(self._message),
                                     b64decode(self._message['Signature']))

    @staticmethod
    def _sign_text(message):
        """
        Reference:
        https://docs.aws.amazon.com/sns/latest/dg/SendMessageToHttp.verify.signature.html
        """
        to_sign = ''

        def sig_pair(el, message):
            return '{}\n{}\n'.format(el, message[el])

        to_sign = sig_pair('Message', message)
        to_sign += sig_pair('MessageId', message)

        if message['Type'] == 'Notification':
            if 'Subject' in message:
                to_sign += sig_pair('Subject', message)
            to_sign += sig_pair('Timestamp', message)

        elif message['Type'] == 'SubscriptionConfirmation':
            to_sign += sig_pair('SubscribeURL', message)
            to_sign += sig_pair('Timestamp', message)
            to_sign += sig_pair('Token', message)

        to_sign += sig_pair('TopicArn', message)
        to_sign += sig_pair('Type', message)

        return to_sign.encode('utf-8')



import logging
from abc import ABC, abstractmethod
from django.conf import settings


class ProcessorException(Exception):
    pass


class MessageBodyProcessor(ABC):
    def __init__(self, logger, queue_settings_name, is_encrypted=False):
        self.logger = logger
        self.settings = settings.AWS_SQS.get(queue_settings_name, {})
        self.is_encrypted = is_encrypted

    def validate_message_body(self, payload):
        """
        Override in the sub-class if the message payload requires validation.
        Must return True or False.
        """
        return True

    def validate_message_body_signature(self, payload):
        """
        Override in the sub-class if the message payload signature requires
        validation.
        """
        pass

    def decrypt_message_body(self, payload):
        """
        Override in the sub-class if the message payload is encrypted
        """
        return payload

    def get_queue_settings(self):
        return self.settings

    def process(self, payload):
        """
        :param payload: the message payload json data
        """
        if self.validate_message_body(payload):

            if self.settings.get('VALIDATE_BODY_SIGNATURE', False):
                self.validate_message_body_signature(payload)

            if self.is_encrypted:
                # the payload is encrypted
                payload = self.decrypt_message_body(payload)

            self.process_message_body(payload)

    @abstractmethod
    def process_message_body(self, payload):
        """
        A sub-class must define this method
        :raises ProcessorException: any error unable to handle
        """
        pass



import boto3
import re
from aws_message.mock_sqs import SQSQueueMock


class SQSException(Exception):
    pass


class SQSQueue(object):

    DEFAULT_WAIT_TIME = 10
    DEFAULT_VISIBILITY_TIMEOUT = 10

    def __init__(self, sqs_settings):
        try:
            self._settings = sqs_settings
            self.arn = self._settings.get('QUEUE_ARN')
            self.key_id = self._settings['KEY_ID']
            self.key = self._settings['KEY']
        except KeyError:
            raise SQSException('Invalid SQS configuration {}'.format(
                self._settings))

        # dig region, account and queue_name out of ARN
        #     arn:aws:sqs:<region>:<account-id>:<queuename>
        # defined at:
        #     https://docs.aws.amazon.com/general/latest/
        #         gr/aws-arns-and-namespaces.html
        m = re.match(r'^arn:aws:sqs:'
                     r'(?P<region>([a-z]{2}-[a-z]+-\d+|mock)):'
                     r'(?P<account_id>\d+):'
                     r'(?P<queue_name>[a-z\d\-\_\.]*)$', self.arn, re.I)
        if not m:
            raise SQSException('Invalid ARN: {}'.format(self.arn))

        self.region = m.group('region')
        self.account_id = m.group('account_id')
        self.queue_name = m.group('queue_name')

        if self.region == 'mock':
            self._queue = SQSQueueMock()
            return

        sqs = boto3.resource(
            'sqs',
            aws_access_key_id=self.key_id,
            aws_secret_access_key=self.key,
            region_name=self.region
        )

        # By default SSL is used
        # By default SSL certificates are verified

        self._queue = sqs.get_queue_by_name(QueueName=self.queue_name)
        if self._queue is None:
            raise SQSException('No queue by name {}'.format(
                self.queue_name))

    def get_messages(self, max_msgs_to_fetch):
        return self._queue.receive_messages(
            AttributeNames=['All'],
            MessageAttributeNames=['All'],
            MaxNumberOfMessages=max_msgs_to_fetch,
            WaitTimeSeconds=self._settings.get(
                'WAIT_TIME', self.DEFAULT_WAIT_TIME),
            VisibilityTimeout=self._settings.get(
                'VISIBILITY_TIMEOUT', self.DEFAULT_VISIBILITY_TIMEOUT))

    def delete_message(self, msg):
        raise SQSException('Invalid method for sqs queue')
        # call msg.delete() would delete the message

    def purge_messages(self):
        """
        Delete all the messages in the queue before purge is called
        """
        return self._queue.purge()



from importlib import import_module
from hashlib import sha1
from http.client import HTTPSConnection
import socket
import ssl
import os
import sys
import inspect
import glob
import re
from django.conf import settings


class HTTPSConnectionValidating(HTTPSConnection):
    def __init__(self, host, port=None, key_file=None,
                 cert_file=None, timeout=None, strict=None):
        HTTPSConnection.__init__(
            self, host, port=port, key_file=key_file,
            cert_file=cert_file, timeout=timeout, strict=strict)
        self.key_file = key_file
        self.cert_file = cert_file
        self.timeout = timeout

    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file,
                                    ca_certs=settings.AWS_CA_BUNDLE,
                                    cert_reqs=ssl.CERT_REQUIRED)


def https_connection_factory(host, port=None, strict=None):
    return HTTPSConnectionValidating(host, port=port, strict=strict,
                                     key_file=settings.EVENT_AWS_SQS_KEY,
                                     cert_file=settings.EVENT_AWS_SQS_CERT)


class SQSQueueMock(object):
    def __init__(self, *args, **kwargs):
        self._settings = kwargs.get('settings')
        self._fs_encoding = sys.getfilesystemencoding() or \
            sys.getdefaultencoding()
        self._app_mock_dirs = []
        self._seen_messages = []
        try:
            for app in settings.INSTALLED_APPS:
                mod = import_module(app)
                mock_dir = os.path.join(
                    os.path.dirname(mod.__file__), 'mock')
                if os.path.isdir(mock_dir):
                    data = {
                        'base': os.path.dirname(
                            mod.__file__).decode(self._fs_encoding),
                        'path': mock_dir.decode(self._fs_encoding),
                        'app': app,
                    }
                    self._app_mock_dirs.append(data)
        except Exception:
            pass

    def get_messages(self, max_msgs_to_fetch):
        try:
            module = self._calling_module()
            mock_dir = self._mock_dir(module)
            msg_files = glob.glob(os.path.join(mock_dir, 'message*'))
            for msg_file in msg_files:
                with open(msg_file, 'r') as handle:
                    msg = handle.read()
                    if self._hash(msg) not in self._seen_messages:
                        return [RawMessage(body=msg)]
        except Exception:
            pass

        return []

    def _hash(self, raw):
        return sha1(raw).hexdigest()

    def delete_message(self, msg):
        hash = self._hash(msg.get_body())
        if hash not in self._seen_messages:
            self._seen_messages.append(hash)

    def _calling_module(self):
        sqs_dir = os.path.dirname(__file__)
        for frame in inspect.stack():
            mod = inspect.getmodule(frame[0])
            if not mod.__file__.startswith(sqs_dir):
                return mod

        raise Exception("no calling module")

    def _mock_dir(self, module):
        dir = os.path.dirname(module.__file__)
        for mock_dir in self._app_mock_dirs:
            if dir.startswith(mock_dir['base']):
                msg_dir = os.path.join(mock_dir['path'],
                                       'aws_message',
                                       self._settings.get('QUEUE'))
                if os.path.isdir(msg_dir):
                    return msg_dir

        raise Exception("no mock directory")



import json
from logging import getLogger
import traceback
from django.conf import settings
from aws_message.crypto import CryptoException
from aws_message.processor import ProcessorException
from aws_message.sqs import SQSQueue
from aws_message.message import Message


logger = getLogger(__name__)


class GatherException(Exception):
    pass


class Gather(object):
    """
    Class to gather event messages from AWS SQS queue,
    validate and process their content
    """

    def __init__(self,
                 processor=None,
                 exception=None,
                 sqs_settings=None):
        """
        :param processor: A sub-class object of MessageBodyProcessor
        """

        if not processor:
            raise GatherException('missing event processor')

        self._processor = processor
        self._settings = sqs_settings if (
            sqs_settings) else self._processor.get_queue_settings()

        self._queue = SQSQueue(self._settings)
        # if Exception, abort!

    def gather_events(self):
        to_fetch = self._settings.get('MESSAGE_GATHER_SIZE')

        while to_fetch > 0:
            messages = self._queue.get_messages(to_fetch)

            if len(messages) == 0:
                logger.info("SQS queue is drained")
                break

            for msg in messages:
                try:
                    # validate the message and hand it off for processing
                    message = Message(json.loads(msg.body), self._settings)
                    if message.validate():
                        self._processor.process(message.extract())

                except (CryptoException, ProcessorException) as err:
                    # log message specific error, abort if unknown error
                    logger.error('{}: {}'.format(
                        err, traceback.format_exc().splitlines()))

                else:
                    # inform the queue that this message has been processed
                    msg.delete()



from django.conf import settings
from django.core.cache import cache
from Crypto.Cipher import AES
from hashlib import sha1
from oscrypto import asymmetric as oscrypto_asymmetric
from oscrypto import errors as oscrypto_errors
import urllib3
import logging

# shunt warnings to logging
logging.captureWarnings(True)


class CryptoException(Exception):
    pass


class Signature(object):
    """
    SHA1 with RSA message signature object
    """

    _cert = None

    def __init__(self, config):
        """
        SHA1 with RSA signed message object

        Takes config object containing certificate reference

        Raises CryptoException
        """
        if config is None:
            raise CryptoException('Missing signature configuration')

        if 'cert' in config:
            cert = config['cert']
        else:
            raise CryptoException('Missing certificate configuration')

        if cert['type'].lower() == 'url':
            cert_ref = cert['reference']
            key = sha1(cert_ref.encode('utf-8')).hexdigest()
            self._cert = cache.get(key)
            if self._cert is None:
                try:
                    http = urllib3.PoolManager(
                        cert_file=config.get('cert_file'),
                        key_file=config.get('key_file'),
                        cert_reqs='CERT_REQUIRED',
                        ca_certs=settings.AWS_CA_BUNDLE
                    )
                    r = http.request('GET', cert_ref)
                    if r.status == 200:
                        self._cert = r.data
                        cache.set(key, self._cert)
                    else:
                        raise CryptoException(
                            'Cannot get certificate {}: status {}'.format(
                                cert_ref or 'None', r.status))
                except urllib3.exceptions.HTTPError as err:
                    raise CryptoException(
                        'Cannot get certificate {}: {}'.format(
                            cert_ref or 'None', err))
        else:
            raise CryptoException('Unrecognized certificate reference type')

    def validate(self, msg, sig):
        if self._cert is None:
            raise CryptoException('Cannot validate: no certificate')

        try:
            oscrypto_asymmetric.rsa_pkcs1v15_verify(
                oscrypto_asymmetric.load_certificate(self._cert),
                sig, msg, 'sha1')
        except oscrypto_errors.SignatureError as err:
            raise CryptoException('Cannot validate: {}'.format(err))


class aes128cbc(object):

    _key = None
    _iv = None

    def __init__(self, key, iv):
        """
        Advanced Encryption Standard object

        Raises CryptoException
        """
        self._block_size = 16

        if key is None:
            raise CryptoException('Missing AES key')
        else:
            self._key = key

        if iv is None:
            raise CryptoException('Missing AES initialization vector')
        else:
            self._iv = iv

    def encrypt(self, msg):
        try:
            crypt = AES.new(self._key, AES.MODE_CBC, self._iv)
            return crypt.encrypt(msg)
        except Exception as err:
            raise CryptoException('Cannot decrypt message: {}'.format(err))

    def decrypt(self, msg):
        try:
            crypt = AES.new(self._key, AES.MODE_CBC, self._iv)
            return crypt.decrypt(msg)
        except Exception as err:
            raise CryptoException('Cannot decrypt message: {}'.format(err))

    def pad(self, s):
        return s + (self._block_size - len(s) % self._block_size) * \
            chr(self._block_size - len(s) % self._block_size)

    def unpad(self, s):
        return s[0:-ord(s[-1])]



import os
from setuptools import setup

README = """
See the README on `GitHub
<https://github.com/uw-it-aca/django-aws-message>`_.
"""

# The VERSION file is created by travis-ci, based on the tag name
version_path = 'aws_message/VERSION'
VERSION = open(os.path.join(os.path.dirname(__file__), version_path)).read()
VERSION = VERSION.replace("\n", "")

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-aws-message',
    version=VERSION,
    packages=['aws_message'],
    include_package_data=True,
    install_requires = [
        'Django>=2.0,<3.0',
        'boto3',
        'pycrypto',
        'oscrypto',
        'urllib3'
    ],
    license='Apache License, Version 2.0',
    description=('An application on which to build AWS SQS endpoints and '
                 'SQS gatherers'),
    long_description=README,
    url='https://github.com/uw-it-aca/django-aws-message',
    author = "UW-IT AXDD",
    author_email = "aca-it@uw.edu",
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
)
